/* tridraw.c - Triangle drawing with planar masking example
 * 2022 John "Lameguy" Wilbert Villamor of Meido-Tek Productions (Lameguy64)
 *
 *
 * This program demonstrates  drawing  of  triangles  under VGA mode 12h with a
 * 640x400 resolution tweak using only fixed-point arithmetic. The example also
 * demonstrates image masking by means of exploiting some unique tricks present
 * only in planar video modes.
 * 
 *
 * Drawing Algorithm:
 *
 *	The triangle drawing implementation uses an all-integer Bresenham algorithm
 *	using 16:16 (65536 = 1.0)  fixed-point  arithmetic with some degree of sub-
 *	pixel calculation. The triangle is  drawn  in two passes split by the point
 *	between the top-most and bottom-most  point  of  the triangle. However, the
 *	implementation in this example does not yet render with 'perfect' sub-pixel
 *	calculation as horizontal edges of the triangle tend to appear as though no
 *	sub-pixel calculation  is  employed  (ie. Lines or polygon routines that do
 *	not take fractional values).
 *
 *	The masking effect is achieved by  exploiting the very nature of the planar
 *	pixel format. By drawing  the  artwork  with  only the first half of the 16
 *	available colors, in this  case  the  first 5, but these colors are matched
 *	such that the drawing appears  as  a solid color while the 'real' colors of
 *	the drawing are placed on the  latter  half of the palette. Setting bits to
 *	the fourth color plane would  shift-up  the color indices of the drawing to
 *	the latter half with the real palette, thus revealing the drawing.
 *
 *	The triangle drawing routine can be  made  to draw solid polygons by simply
 *	invoking it while in VGA write mode 3.  Planes  to  modify and color of the
 *	triangle are  defined  by  the  color  plane  write  enable  and  set/reset
 *	registers.
 *
 *
 * Compiling:
 *
 *	This example can be compiled  using  either  OpenWatcom or DJGPP. Compiling
 *	with the original Watcom C  compiler  is  not  tested. Supporting both com-
 *	pilers was the result of  testing  optimized  code  performance between the
 *	two compilers (spoiler: DJGPP is faster than OpenWatcom).
 *
 *	To compile using OpenWatcom simply invoke 'wmake' using the file
 *	'makefile.wat':
 *
 *		wmake -f makefile.wat
 *
 *	This should generate programs  gentable.exe and demo_wat.exe, the former of
 *	which generates the tables.dat  file.  gentable.exe  is  compiled using the
 *	host system's C compiler to guarantee  successful  compiling under a 64-bit
 *	Windows environment.  The same should occur when compiling with DJGPP under
 *	Linux.
 *
 *	To compile using DJGPP simply invoke 'make', 'gmake' or 'gnumake' using the
 *	file 'makefile.gcc':
 *
 *		make -f makefile.gcc
 *
 *	The makefile assumes the DJGPP tool-chain possesses a djgpp-* suffix. If no
 *	suffix is present (ie.  Using pre-compiled builds of DJGPP for MS-DOS) or a
 *	different suffix is used  altogether  (ie. Cross-compiling from Linux), the
 *	C compiler command can be overridden with CC=<compiler-name>:
 *
 *		make -f makefile.gcc CC=gcc
 *
 *		or
 *
 *		make -f makefile.gcc CC=i586-pc-msdosdjgpp-gcc
 *
 */
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include "pcx.h"

/*
 * Program defines
 */

/* Maximum number of polygons to display */
#define MAX_POLYS	8


/*
 * Program Structs
 */

/* Struct for x,y coordinates */
typedef struct _VECTOR
{
	int x;
	int y;
} VECTOR;

/* Drawing stage struct for DrawTriangle */
typedef struct _DRAWSTG
{
	int nrows;
	int xdelta;
	int xx;
	int pad;
} DRAWSTG;

/* Moving polygon entry */
typedef struct _POLYENTRY
{
	int x,y;
	int ang;
	int xvel,yvel;
	int angvel;
	short points;
} POLYENTRY;

/* Look-up table arrays */
int stable[4096];
int ctable[4096];

/* Memory offset to specify which video page to draw to */
int page_offs;


/*
 * VGA stuff
 */

#define VGA_SEGMENT			0xA0000	/* VGA memory segment */
#define VGA_WIDTH			80		/* Number of bytes per scanline */

#define _BIOS_SET_MODE		0x0000	/* BIOS video mode set */
#define _BIOS_SET_PALETTE	0x1000	/* BIOS video palette set  */

#ifdef __WATCOMC__					/* (Open)Watcom specific defines */

/* Types compatibility for (Open)Watcom */
#if !defined(uint8_t)
typedef unsigned char	uint8_t;
#endif
#if !defined(uint16_t)
typedef unsigned short	uint16_t;
#endif

/* Macro for invoking the VGA interrupt vector */
extern short VideoInt(short, short, short, short);

#pragma aux	VideoInt = \
			"push	bp		", \
			"int	10h		", \
			"pop	bp		", \
			parm caller [ax] [bx] [cx] [dx] value [ax];

/* To save having to place a load of #ifdefs within graphics functions */

#define _farnspokeb(addr, data)	( *((uint8_t*)addr) = data )
#define _farnspokew(addr, data) ( *((uint16_t*)addr) = data )
#define _farnspokel(addr, data)	( *((uint32_t*)addr) = data )

#define _farnspeekb(addr)		( *((uint8_t*)addr) )
#define _farnspeekw(addr)		( *((uint16_t*)addr) )
#define _farnspeekl(addr)		( *((uint32_t*)addr) )

#else								/* DJGPP specific defines */

#include <sys/farptr.h>
#include <go32.h>

/* Don't know how to implement this as an asm macro yet */
short VideoInt(short ax, short bx, short cx, short dx)
{
	union REGS regs;

	regs.x.ax = ax;
	regs.x.bx = bx;
	regs.x.cx = cx;
	regs.x.dx = dx;
	
	int386(0x10, &regs, &regs);
	
	return regs.x.ax;

} /* VideoInt */

#endif /* __WATCOMC__ */


/* initVGA - Sets VGA mode to 12h with a 640x400 resolution tweak
 *
 * Arguments:
 *	None
 *
 * Returns:
 *	None
 *
 */
void initVGA(void)
{
	int i;
	
	/* Set video mode 12h */
	VideoInt(_BIOS_SET_MODE+0x12, 0, 0, 0);

	/* Tweak mode for 640x400 resolution at 70Hz
	 * (based on Michael Abrash's tweaks)
	 */
	outp(0x3d4, 0x11); outp(0x3d5, inp(0x3d5)&0x7f);
	outp(0x3c2, (inp(0x3cc)&0x3f)|0x40);
	outpw(0x3d4, 0xbf06);
	outpw(0x3d4, 0x1f07);
	outpw(0x3d4, 0x4009);
	outpw(0x3d4, 0x9c10);
	outpw(0x3d4, 0x8e11);
	outpw(0x3d4, 0x8f12);
	outpw(0x3d4, 0x9615);
	outpw(0x3d4, 0xb916);
	outp(0x3d4, 0x11); outp(0x3d5, inp(0x3d5)|0x80);
	
	/* Remap VGA palette so colors can be changed */
	VideoInt(_BIOS_SET_PALETTE+0x01, 16<<8, 0, 0);
	for(i=0; i<16; i++)
		VideoInt(_BIOS_SET_PALETTE+0x00, (i<<8)|i, 0, 0);
	
	/* Map overscan color to index 16, so 17 unique colors are possible :) */
	VideoInt(_BIOS_SET_PALETTE+0x10, 16, 0, 0);
	
} /* initVGA */


/* initText - Sets VGA text mode
 *
 * Arguments:
 *	None
 *
 * Returns:
 *	None
 *
 */
void initText(void)
{
	/* Set text mode */
	VideoInt(_BIOS_SET_MODE+0x03, 0, 0, 0);
	
} /* initText */


/* setScanline - Sets the video memory display offset in scanline units
 *
 * Arguments:
 *	line		Starting scanline to display
 *
 * Returns:
 *	None
 *
 */
void setScanline(int line)
{
	line *= VGA_WIDTH;

	outpw(0x3d4, 0x000c|(line&0xff00));
	outpw(0x3d4, 0x000d|((line&0xff)<<8));
	
} /* setScanline */


/* waitVSync - Waits for a state change of the vertical retrace bit
 *
 * Arguments:
 *	mode		0:Wait until start of retrace, 1:Wait until end of retrace
 *
 * Returns:
 *	None
 *
 */
void waitVSync(int mode)
{
	if( mode == 0 )
	{
		while( (inp(0x3da)&0x8) );
	}
	else
	{
		while( !(inp(0x3da)&0x8) );
	}

} /* waitVSync */


/* loadTable - Loads look-up table data from the file 'tables.dat'
 *
 * Arguments:
 *	None
 *
 * Returns:
 *	0		Table loaded OK
 *	1		tables.dat file not found
 *
 */
int loadTable(void)
{
	FILE *fp;

	fp = fopen("tables.dat", "rb");
	if( !fp )
		return 1;

	fread(stable, 4, 4096, fp);
	fread(ctable, 4, 4096, fp);

	fclose(fp);

	return 0;
	
} /* loadTable */


/* DrawTriangle - Draws a masking triangle with some sub-pixel calculation
 *
 * Arguments:
 *	x1,x2,x3	X coordinates of triangle (32768 = 1.0 pixel)
 *	y1,y2,y3	Y coordinates of triangle (32768 = 1.0 pixel)
 *
 * Returns:
 *	None
 *
 */
void DrawTriangle(int x1, int y1, int x2, int y2, int x3, int y3)
{
	DRAWSTG stage[2];

	int i,j,k,l,step;
	int xx;
	int xdelta;
	int topclip,btmclip;
#ifdef __WATCOMC__
	uint8_t *vptr,*vtmp;
#else
	uint32_t vptr,vtmp;
#endif
	
	/* Bits table for the left-most tip of a row */
	static uint8_t llut[8] =
	{
		0xff,0x7f,0x3f,0x1f,
		0x0f,0x07,0x03,0x01
	};

	/* Bits table for the right-most tip of a row */
	static uint8_t rlut[8] =
	{
		0x00,0x80,0xc0,0xe0,
		0xf0,0xf8,0xfc,0xfe
	};

	/* Make sure vertices are oriented top to bottom */
	if( y1 > y3 )
	{
		i = x1; x1 = x3; x3 = i;
		i = y1; y1 = y3; y3 = i;
	}

	if( y2 > y3 )
	{
		i = x2; x2 = x3; x3 = i;
		i = y2; y2 = y3; y3 = i;
	}

	if( y1 > y2 )
	{
		i = x1; x1 = x2; x2 = i;
		i = y1; y1 = y2; y2 = i;
	}

	/* Compute total number of rows of triangle and X delta */
	stage[1].nrows = (y3>>16)-(y1>>16);
	if( stage[1].nrows )
		xdelta = (x3-x1)/stage[1].nrows;
	else
		xdelta = 0;

	/* Do the same for the top-half of triangle */
	stage[1].nrows = (y2>>16)-(y1>>16);
	if( stage[1].nrows )
		stage[1].xdelta = (x2-x1)/stage[1].nrows;
	else
		stage[1].xdelta = 0;

	/* Then bottom-half of triangle */
	stage[0].nrows = (y3>>16)-(y2>>16);
	if( stage[0].nrows )
		stage[0].xdelta = (x3-x2)/stage[0].nrows;
	else
		stage[0].xdelta = 0;

	/* Vertical clipping of top-half */
	topclip = 0;
	if( y1 < 0 )
	{
		topclip = 0-(y1>>16);
		y1 = 0;
	}
	
	btmclip = 0;
	if( (y2>>16) > 400 )
	{
		btmclip = (y2>>16)-400;
	}

	/* Adjust number of rows and initial X coordinate after clipping */
	stage[1].nrows -= (topclip+btmclip);
	if( stage[1].nrows < 0 )
		stage[1].nrows = 0;

	xx = x1+(xdelta*topclip);
	stage[1].xx = x1+(stage[1].xdelta*topclip);
		
	/* Vertical clipping of bottom-half */
	topclip = 0;
	if( y2 < 0 )
	{
		topclip = 0-(y2>>16);
		y2 = 0;
	}
	
	btmclip = 0;
	if( (y3>>16) > 400 )
	{
		btmclip = (y3>>16)-400;
	}

	/* Adjust number of rows and initial X coordinate after clipping */
	stage[0].nrows -= (topclip+btmclip);
	if( stage[0].nrows < 0 )
		stage[0].nrows = 0;
		
	stage[0].xx = x2+(stage[0].xdelta*topclip);

	/* Calculate initial video memory address to draw pixels to */
#ifdef __WATCOMC__
	vptr = ((uint8_t*)VGA_SEGMENT+VGA_WIDTH*(y1>>16))+page_offs;
#else
	vptr = ((uint32_t)VGA_SEGMENT+VGA_WIDTH*(y1>>16))+page_offs;
	_farsetsel(_dos_ds);
#endif

	/* Two-pass drawing procedure */
	step = 2;
	while( step )
	{
		step--;

		/* Draw a half of the triangle line by line */
		for(i=0; i<stage[step].nrows; i++)
		{
			/* Swap coordinates if not Left-to-Right */
			if( xx > stage[step].xx )
			{
				k = stage[step].xx>>16;
				l = xx>>16;
			}
			else
			{
				k = xx>>16;
				l = stage[step].xx>>16;
			}

			/* Reject line if both coordinates are off-screen */
			if( ( k < 0 ) && ( l < 0 ) )
			{
				vptr += VGA_WIDTH;
				xx += xdelta;
				stage[step].xx += stage[step].xdelta;
				continue;
			}

			if( ( k > 640 ) && ( l > 640 ) )
			{
				vptr += VGA_WIDTH;
				xx += xdelta;
				stage[step].xx += stage[step].xdelta;
				continue;
			}
			
			/* Perform X clipping per line (sadly) */
			if( k < 0 )
				k = 0;

			if( l > 640 )
				l = 640;

			/* Apply line offset to framebuffer pointer */
			vtmp = vptr+(k>>3);

			if( (j = l-k) > 7 )		/* if line is more than 7 pixels long */
			{
				/* Calculate width */
				j = (l>>3)-(k>>3);
				
				/* Draw start cap */
				if( k&0x7 )
				{
					//*vtmp |= llut[k&0x7];
					_farnspokeb(vtmp, _farnspeekb(vtmp)|llut[k&0x7]);
					vtmp++;
					j--;
				}
				
				/* Pre-fill to align pointer for 32-bit writes */
				while( ((unsigned int)vtmp&0x3) && (j>0) )
				{
					//*vtmp = 0xff;
					_farnspokeb(vtmp, 0xff);
					vtmp++;
					j--;
				}
				
				/* Fill span with 32-bit writes */
				while(j > 4)
				{
					//*((unsigned int*)vtmp) = 0xffffffff;
					_farnspokel(vtmp, 0xffffffff);
					vtmp += 4;
					j -= 4;
				}

				/* Fill the remainder */
				while( j>0 )
				{
					//*vtmp = 0xff;
					_farnspokeb(vtmp, 0xff);
					vtmp++;
					j--;
				}
				
				/* Draw end cap */
				if( l&0x7 )
				{
					//*vtmp |= rlut[l&0x7];
					_farnspokeb(vtmp, _farnspeekb(vtmp)|rlut[l&0x7]);
					vtmp++;
				}
			}
			else					/* when line is less than 7 pixels */
			{
				if( (k&0x7)+j > 7 )
				{
					//*vtmp |= llut[k&7]; vtmp++;
					//*vtmp |= rlut[l&7];
					_farnspokeb(vtmp, _farnspeekb(vtmp)|llut[k&7]); vtmp++;
					_farnspokeb(vtmp, _farnspeekb(vtmp)|rlut[l&7]);
				}
				else
				{
					//*vtmp |= ((1<<j)-1)<<((8-j)-(k&0x7));
					_farnspokeb(vtmp, _farnspeekb(vtmp)|
						((1<<j)-1)<<((8-j)-(k&0x7)));
				}
			}

			vptr += VGA_WIDTH;
			xx += xdelta;
			stage[step].xx += stage[step].xdelta;
		}
	}
	
} /* DrawTriangle */


/* clearScreen - Fills 32,000 zero-value bytes to page_offs
 *
 * Used to clear the screen if the VGA write plane register is set to all
 * video planes or specific video plane(s).
 *
 * Arguments:
 *	None
 *
 * Returns:
 *	None
 *
 */
void clearScreen(void)
{
	int i;
#ifdef __WATCOMC__
	uint8_t *vptr = (uint8_t*)VGA_SEGMENT+page_offs;
#else
	uint32_t vptr = (uint32_t)VGA_SEGMENT+page_offs;
	_farsetsel(_dos_ds);
#endif

	for(i=0; i<32000; i++)
	{
		_farnspokeb(vptr, 0);
		vptr++;
	}
	
} /* clearScreen */


/* PutImage - Copies a full-screen planar image directly to video memory
 *
 * Source data is assumed to be four copies of 1-bit bitmaps, one per plane.
 * The order of planes are plane0:plane1:plane2:plane3 without any form of
 * interleaving. The bitmap data of each plane can be interpreted as 1-bit
 * per pixel images.
 *
 * Arguments:
 *	*data		Pointer to raw planar bitmap data
 *	width		Width of image in byte increments
 *	height		Height of image in scanlines
 *
 * Returns:
 *	None
 *
 */
void PutImage(uint8_t *data, int width, int height)
{
    int p,px,py;
#ifdef __WATCOMC__
    uint8_t *vptr;
#else
    uint32_t vptr;
    _farsetsel(_dos_ds);
#endif

    for(p=0; p<4; p++)
    {
		outp(0x3c4, 0x02);
		outp(0x3c5, 1<<p);
#ifdef __WATCOMC__
		vptr = ((uint8_t*)VGA_SEGMENT)+page_offs;
#else
		vptr = ((uint32_t)VGA_SEGMENT)+page_offs;
#endif
		for(py=0; py<height; py++)
		{
			for(px=0; px<width; px++)
			{
				_farnspokeb(vptr, *data);	
				vptr++;
				data++;
			}
		}
	}
	
} /* PutImage */


/* drawPoly - Draws a polygon entry
 *
 * Arguments:
 *  *entry		Pointer to a POLYENTRY item
 *
 * Returns:
 *	None
 *
 */
void drawPoly(POLYENTRY *entry)
{
	VECTOR v[5];
   
	int i,j;
	int vx,vy,ang;

	j = 0;
	ang = entry->ang&4095;
	for(i=0; i<entry->points; i++)
	{
		vx = 75*stable[j&4095];
		vy = 75*ctable[j&4095];

		v[i].x =
			(((long long)vx*ctable[ang])>>16)-
			(((long long)vy*stable[ang])>>16);
		v[i].y =
			(((long long)vy*ctable[ang])>>16)+
			(((long long)vx*stable[ang])>>16);
		v[i].x = entry->x+v[i].x;
		v[i].y = entry->y+v[i].y;
		//v[i].x = x+(((long long)v[i].x*scale)>>16);
		//v[i].y = y+(((long long)v[i].y*scale)>>16);

		j += 4096/entry->points;
	}

	i = 0;
	while(i < entry->points-2)
	{	
		DrawTriangle(
			v[0].x, v[0].y,
			v[i+1].x, v[i+1].y,
			v[i+2].x, v[i+2].y);
		i++;
	}


} /* drawPoly */


/* musicRoom - Music Room screen
 *
 * A sort-of Touhou inspired 'Music Room' demonstrating the DrawTriangle()
 * function to achieve a masking effect, where a bitmap appears to be drawn
 * through the polygon.
 *
 * Arguments:
 *	None
 *
 * Returns:
 *	None
 *
 */
void musicRoom(void)
{
	POLYENTRY polylist[MAX_POLYS];
	
	int i;
	int page;

	// Generate initial parameters of polygon entries
	for(i=0; i<MAX_POLYS; i++)
	{
		polylist[i].x = (rand()%640)<<16;
		polylist[i].y = -(100<<16);
		polylist[i].xvel = (-16383+(rand()&32767))<<3;
		polylist[i].yvel = 65536+((rand()&32767)<<2);
		polylist[i].ang = rand()&4095;
		polylist[i].angvel = (-64+(rand()&127));
		polylist[i].points = 3+(rand()%3);
	}

	page = 0;
	while( !kbhit() )
	{
		page_offs = (80*400)*page;
		outp(0x3c4, 0x02); outp(0x3c5, 0x4);
		outp(0x3ce, 0x04); outp(0x3cf, 2);
		
		clearScreen();
		
		for(i=0; i<MAX_POLYS; i++)
		{
			drawPoly(&polylist[i]);

			polylist[i].x += polylist[i].xvel;
			polylist[i].y += polylist[i].yvel;
			polylist[i].ang += polylist[i].angvel;

			if( polylist[i].x < 0 )
			{
				polylist[i].xvel = -polylist[i].xvel;
				polylist[i].angvel = -polylist[i].angvel;
			}
			if( (polylist[i].x>>16) > 640 )
			{
				polylist[i].xvel = -polylist[i].xvel;
				polylist[i].angvel = -polylist[i].angvel;
			}

			if( (polylist[i].y>>16) > 500 )
			{
				polylist[i].y = -(100<<16);
				polylist[i].xvel = (-16383+(rand()&32767))<<3;
				polylist[i].yvel = 65536+((rand()&32767)<<2);
				polylist[i].ang = rand()&4095;
				polylist[i].angvel = (-64+(rand()&127));
				polylist[i].points = 3+(rand()%3);
			}
		}

		setScanline(400*page);
		waitVSync(0);
		waitVSync(1);

		page = !page;
	}

} /* musicRoom */


int main(int argc, const char *argv[])
{
	int i;

	uint8_t *logobg_data;
	uint8_t logobg_pal[48];
	int logobg_w,logobg_h;
	
	argc = argc;
	argv = argv;

	/* Load sin/cos tables */
	if( loadTable() )
	{
		printf("I cannot load tables.dat!\n");
		return 1;
	}

	/* Load the artwork */
	if( ( logobg_data = loadpcx("backgrnd.pcx",
		&logobg_w, &logobg_h, logobg_pal) ) == NULL )
	{
		printf("I cannot load backgrnd.pcx!\n");
		return 1;
	}

	/* Set graphics mode */
	initVGA();

	/* Set artwork palette */
	outp(0x3c8, 0);
	for(i=0; i<48; i+=3)
	{
		outp(0x3c9, logobg_pal[i]);
		outp(0x3c9, logobg_pal[i+1]);
		outp(0x3c9, logobg_pal[i+2]);
	}

	/* Load the artwork to both pages */
	page_offs = 0;
	PutImage(logobg_data, logobg_w, logobg_h);
	page_offs = 80*400;
	PutImage(logobg_data, logobg_w, logobg_h);

	/* Free the artwork data once it's loaded into video memory */
	free(logobg_data);

	/* Enter the 'Music Room' screen, even though there's no music */
	musicRoom();
	
	/* To flush the key press from exiting the 'Music Room' */
	getch();

	/* Return to text mode */
	initText();
	
	return 0;
	
} /* main */
